/*------------------------------------------------------------------------/
/  Experimental SD interface mode (not SPI) control module
/-------------------------------------------------------------------------/
/
/  Copyright (C) 2024, Evan Hillas, all right reserved.
/
/ * This software is a free software and there is NO WARRANTY.
/ * No restriction on use. You can use, modify and redistribute it for
/   personal, non-profit or commercial products UNDER YOUR RESPONSIBILITY.
/ * Redistributions of source code must retain the above copyright notice.
/
/-------------------------------------------------------------------------/
    v0.5, 25 Nov 2024

  Features and Limitations:

  * Platform Specific - Prop2 - for performance
    Uses hardware DMA resources and "smartpin"s

  * No Media Change Detection
    Application program needs to perform a f_umount()/f_mount() after media change.

  * Uses 6, 7 or 8 I/O pins, 4-bit mode only
    DAT pins must be ordered on a 4-bit pin boundary
      Eg: Pins P0..3 or P4..7 or P8..11 and so on

/-------------------------------------------------------------------------*/

#include "../fatfs/diskio.h"    // Common include file for FatFs and disk I/O layer
#include <stdlib.h>    // needed for some data types
#include <string.h>    // needed for memset()

#ifndef __propeller2__
#error ONLY SUPPORTS THE PROPELLER 2 MICROCONTROLLER
#endif
#include <propeller2.h>


//--------------------------------------------------------------------------
// Compile options
//--------------------------------------------------------------------------

#ifndef NOREADBLOCKCRC
#define SD_BLOCK_CRC    // enables block read CRC checking
#endif
//#define SD_PR_PARAMS
//#define SD_USE_ACMD23
#define SD_DEBUG
//#define SD_DEBUG_PERFORMANCE


enum {
    RX_SCHMITT = 0,    // 0 = CMOS threshold (CLK pin is Schmitt), 1 = Schmitt trigger of CMD and DAT pins
//    CMDAT_REG = 0,    // 0 = unregistered pins, 1 = registered CMD and DAT pins
    CLK_REG = 0,    // 0 = unregistered pin, 1 = registered CLK pin
    CLK_POL = 0,    // 0 = idle low (selects High-Speed), 1 = idle high (selects Default-Speed)
//    CLK_DIV = 3,    // minimum of 2 (SD clock divider of sysclock), 3 when BLOCK_CRC defined

    COGCRCBUFF = 126,    // cogRAM address of CRC processing buffer (130 longwords size)
#ifdef SD_PR_PARAMS
    SET_PARAMS = 0x1e0,
#else
    SET_PARAMS = COGCRCBUFF - 8,
#endif
};



//--------------------------------------------------------------------------
//
//   Module Private Functions
//
//--------------------------------------------------------------------------


// SD Card/Socket specific variables
uint8_t  cid[18];
uint16_t  rca16;    // active card's 16-bit RCA register
uint16_t  clkdivider;    // full speed clock divider
uint8_t  rxlagcomp;    // Compensates for cumulative clk-rx latencies,
                                // recalibration triggers on read errors
DSTATUS  Stat;    // Disk status
uint8_t  clkpin;    // pin number of clock pin
uint8_t  cmdpin;    // pin number of command/response pin
uint8_t  dat0pin;    // base pin number of the four DAT pins
uint8_t  powpin;    // pin number of power switch (-1 == not used)
uint16_t  ledpin;    // pin number of activity LED (-1 == not used),
                     // doubles up as SDSC byte addressing flag in bit0,
                     // tripples up as SDHC caching flag in bit1,




//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
static void  releasepins( void )
{
    unsigned  PIN_CLK = clkpin;
    unsigned  PIN_CMD = cmdpin;
    unsigned  PIN_DAT = dat0pin | 3<<6;
    unsigned  PIN_POW = powpin;
    unsigned  PIN_LED = ledpin >> 8;    // Doubles up as SDSC byte addressing flag in bit0

    if( PIN_POW != PIN_CLK ) {    // power switch is present
        _pinf(PIN_POW);
        _wrpin(PIN_POW, 0);
    }
    if( PIN_LED != PIN_CLK ) {    // activity LED is present
        _pinf(PIN_LED);
        _wrpin(PIN_LED, 0);
    }

    _pinf(PIN_CLK);    // stop SD clock

    _pinf(PIN_CMD);
    _pinf(PIN_DAT);
    _wrpin(PIN_CMD, 0);
    _wrpin(PIN_DAT, 0);

    _waitus(1);    // wait for pull-up resistors to act

    _wrpin(PIN_CLK, 0);    // release clock pin

    _waitus(1);    // wait for pull-up to act
}



//-----------------------------------------------------------------------
//-----------------------------------------------------------------------

static int  sdcard_power( void )
{
    unsigned  PIN_VOLT = clkpin;
    unsigned  PIN_CMD = cmdpin;
    unsigned  PIN_DAT = dat0pin | 3<<6;
    unsigned  PIN_POW = powpin;    // pin number of power switch (-1 == not used)
    uint32_t  samples, threshold;
    int32_t  tmr, timeout;

    releasepins();    // SD spec 6.4.2

    _pinl(PIN_CMD);
    _pinl(PIN_DAT);
    _pinf(PIN_CMD);
    _pinf(PIN_DAT);
    _waitus(1);    // wait for pull-up resistors to act

    if( PIN_POW == PIN_VOLT ) {    // power switch not present
        if( _pinr(PIN_CMD) )
            if( _pinread(PIN_DAT) == 15 )
                return 1;    // pull-up resistors are present
#ifdef SD_DEBUG
        __builtin_printf(" Error: CMD/DAT pull-up resistors are missing\n");
#endif
        return 0;
    }

    if( _pinr(PIN_POW) ) {    // An inserted card will pull this down against the 22k ohm pull-up
#ifdef SD_DEBUG
        __builtin_printf(" Error: SD Card not detected\n");
#endif
        return 0;
    }
#ifdef SD_DEBUG
    __builtin_printf(" Card detected ... power cycle of SD card\n");
#endif

// Use CompDAC pin mode to wait for supply to fall below 0.5 Volt, SD spec 6.4.1.2
    threshold = (int)(0.5 / 3.4 * 255.0);    // 0.5 Volts
    timeout = 500 * 32;    // 500 ms
#ifdef SD_DEBUG
    __builtin_printf("  power-down threshold = %d   pin state = %d\n", threshold, _pinr(PIN_VOLT));
#endif
    _wrpin(PIN_VOLT, P_LEVEL_A | threshold<<8);    // set CompDAC threshold to 0.5 volts
    _pinh(PIN_POW);    // power off the SD card by overpowering the 1800 ohm Card Detect
    tmr = _cnt();
    _pinl(PIN_DAT);    // drive DAT pin group low to discharge the now unpowered capacitors
    _pinl(PIN_CMD);    // ditto CMD
    _waitus(1);    // DAC settle time is about 1.0 us
    samples = _pinr(PIN_VOLT);
    do {
        _waitus(32);    // 32 us x 32 samples = 1.0 ms
        samples = samples<<1 | _pinr(PIN_VOLT);
        timeout--;
    } while( samples && timeout );    // unanimous wait for pin to drop below threshold level
    tmr = _cnt() - tmr;

    if( samples ) {
#ifdef SD_DEBUG
        __builtin_printf("  Error: Stuck high at SD power-down\n");
#endif
        releasepins();
        return 0;
    }
#ifdef SD_DEBUG
    tmr = _muldiv64(tmr, 1_000_000, _clockfreq());
    __builtin_printf("  power-down slope = %d us   pin state = %d\n", tmr, _pinr(PIN_VOLT));
#endif
//    _waitms(50);

    // Use CompDAC pin mode to wait for supply to rise above 2.7 Volts, SD spec 6.4.1.3
    threshold = (int)(2.7 / 3.3 * 255.0 + 0.5);    // 2.7 Volts
    timeout = 36 * 32;    // 35 ms + 1 ms to cover filter effect
#ifdef SD_DEBUG
    __builtin_printf("  power-up threshold = %d   pin state = %d\n", threshold, _pinr(PIN_VOLT));
#endif
    _pinf(PIN_DAT);
    _pinf(PIN_CMD);
    _pinf(PIN_POW);    // and power back on, released back to Card Detect function
    tmr = _cnt();
    _wrpin(PIN_VOLT, P_LEVEL_A | threshold<<8);    // set CompDAC threshold to 2.7 volts
    _waitus(1);    // DAC settle time is about 1.0 us
    samples = !_pinr(PIN_VOLT);
    do {
        _waitus(32);    // 32 us x 32 samples = 1.0 ms
        samples = samples<<1 | !_pinr(PIN_VOLT);
        timeout--;
    } while( samples && timeout );    // unanimous wait for pin to rise above threshold level
    tmr = _cnt() - tmr;
    if( samples ) {
#ifdef SD_DEBUG
        __builtin_printf("  Error: Stuck low at SD power-up\n");
#endif
        releasepins();
        return 0;
    }
#ifdef SD_DEBUG
    tmr = _muldiv64(tmr, 1_000_000, _clockfreq());
    __builtin_printf("  power-up slope = %d us   pin state = %d\n", tmr, _pinr(PIN_VOLT));
#endif

    releasepins();

    return 1;
}



//-----------------------------------------------------------------------
//-----------------------------------------------------------------------

static uint32_t  crc7sd(    // SD spec 4.5
    uint8_t *buf,
    size_t len )
{
    uint32_t  crc = 0;
    uint32_t  val;

    __asm const {    // "const" enforces XIP, "volatile" enforces Fcache
// Reference code courtesy of Ariba
crc7lp
		rdlong	val, buf
		add	buf, #4
		movbyts	val, #0b00_01_10_11    // byte swap within longwords
		setq	val
		crcnib	crc, #0x48    // CCITT 7-bit polynomial is x0 + x3 + x7
		crcnib	crc, #0x48    // 0x09 reversed and shifted for CRCNIB
		djz	len, #crc7done
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		djz	len, #crc7done
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		djz	len, #crc7done
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		djnz	len, #crc7lp
crc7done
		rev	crc    // correct the bit order to match standard
		shr	crc, #24
		or	crc, #1    // add the SD response end-bit as 8th bit
    }
    return crc;
}



//-----------------------------------------------------------------------
// Tx data block's DMA parameter setup
//-----------------------------------------------------------------------

static struct txblk_parms_t {
    uint32_t  p_clk, p_dat, m_leadin, m_nco, m_se2, m_dat, m_crc, m_leadin2;
} txblkset;

enum {
      r_clk = SET_PARAMS, r_pdat, r_leadin, r_nco, r_se2, r_mdat, r_crc, r_leadin2
};


static void  set_txblkset(
    uint32_t clkdiv,
    uint32_t CMDAT_REG )
{
    unsigned  PIN_DAT0 = dat0pin;
    CMDAT_REG = CMDAT_REG & 1;    // extract lsb from rxlag

    txblkset.p_clk = clkpin;
    txblkset.p_dat = PIN_DAT0 | 3<<6;
    txblkset.m_se2 = 0b110<<6 | PIN_DAT0;    // trigger on high level - card ready
    txblkset.m_leadin = X_IMM_32X1_1DAC1 | 5 + CLK_REG - CMDAT_REG + clkdiv + (CLK_POL ? clkdiv : clkdiv/2);  // start(S)-bit is ahead of the first streamer bit
    txblkset.m_leadin2 = X_IMM_32X1_1DAC1 | 5 + CLK_REG - CMDAT_REG + (CLK_POL ? clkdiv : clkdiv/2);
    txblkset.m_nco = 0x8000_0000UL / clkdiv + (0x8000_0000UL % clkdiv > 0 ? 1 : 0);  // data transfer rate, round up upon a non-zero remainder
    txblkset.m_dat = X_RFBYTE_4P_1DAC4 | PIN_DAT0<<17 | X_PINS_ON | X_ALT_ON | 512 * 2;  // mode and nibble count
    txblkset.m_crc = X_IMM_8X4_1DAC4 | PIN_DAT0<<17 | X_PINS_ON | 8;  // mode and nibble count
}


//-----------------------------------------------------------------------
// Transmit a data block to the card
//-----------------------------------------------------------------------

static int  tx_datablock(    // $3F = no response, $00 = busy timeout, %10_101_1 = CRC mismatch, %10_010_1 = SD card happy
    const uint8_t *buf,    // Buffer address of block data to be transmitted
    uint32_t timeout )    // 250 ms timeout for card busy, in sysclock ticks
{
    __asm const {    // "const" enforces XIP, "volatile" enforces Fcache
// retrieve parameter set
		loc	pa, #txblkset    // fast copy from struct
		setq	#sizeof(txblkset) / 4 - 1
		rdlong	r_clk-0, pa    // fast copy to cogRAM
// start clocking to see when card exits from write busy status
		wypin	##-1, r_clk-0    // busy check
		getct	pb    // timeout begins
		add	timeout, pb    // SDHC 250 ms timeout of block erase-write (SD spec fig 4-35)

		setse2	r_se2-0    // trigger on high level - card ready
// preload block data into CogRAM for CRC processing
		setq	#512/4-1    // one SD block + CRC
		rdlong	COGCRCBUFF, buf    // fast copy to cogRAM before engaging the FIFO

		pollse1       // clear stale event
		pollse2       // clear stale event
    }

    __asm volatile {    // "const" enforces XIP, "volatile" enforces Fcache
		rdfast	lnco, buf    // start the FIFO
		setxfrq	lnco
		setq	timeout    // apply the 250 ms timeout to WAITSEn
		waitse2   wz    // wait for ready (DAT0 high), Z set if timed-out - still busy
		wypin	#2, r_clk-0    // busy turnaround, Nwr (SD spec fig 4-32)
		waitse1

		dirl	r_clk-0    // reset smartpin
	if_nz	drvl	r_pdat-0    // start-bit placed on data pins

	if_nz	xinit	r_leadin-0, #0    // lead-in delay from here at sysclock/1
	if_nz	setq	r_nco-0        // streamer transfer rate (takes effect with buffered command below)
	if_nz	xzero	r_mdat-0, #0     // rx buffered-op, aligned with clock via lead-in
		dirh	r_clk-0    // clock timing starts here
	if_nz	wypin	clocks, r_clk-0    // first pulse outputs during second clock period

// verify the CRC
	if_nz	rep	@.rend1, #512/4    // one SD block + CRC
	if_nz	alti	ptr, #0b100_111_000    // next D-field substitution, then increment ptr, result goes to PA
	if_nz	movbyts	pa, #0b00_01_10_11    // byte swap within longword
	if_nz	splitb	pa    // 8-nibble order swapped to bit order of 4 bytes
	if_nz	setq	pa
	if_nz	crcnib	crc3, poly
	if_nz	crcnib	crc3, poly
	if_nz	crcnib	crc2, poly
	if_nz	crcnib	crc2, poly
	if_nz	crcnib	crc1, poly
	if_nz	crcnib	crc1, poly
	if_nz	crcnib	crc0, poly
	if_nz	crcnib	crc0, poly
.rend1
	if_nz	rolbyte	pa, crc3, #0
	if_nz	rolbyte	pa, crc2, #0
	if_nz	rolbyte	pa, crc1, #0
	if_nz	rolbyte	pa, crc0, #0
	if_nz	mergeb	pa    // 4-byte order swapped to bit order of 8 nibbles
	if_nz	waitse1

	if_nz	setxfrq	lnco
	if_nz	dirl	r_clk-0    // reset smartpin
	if_nz	xinit	r_leadin2-0, #0    // lead-in delay from here at sysclock/1
	if_nz	setq	r_nco-0        // streamer transfer rate (takes effect with buffered command below)
	if_nz	xzero	r_crc-0, pa    // 8 nibbles, transmit low nibble first
	if_nz	dirh	r_clk-0    // clock timing starts here
	if_nz	wypin	#16+2+CLK_POL, r_clk-0    // 16-bit CRC + end-bit
	if_nz	rolbyte	pa, crc3, #1
	if_nz	rolbyte	pa, crc2, #1
	if_nz	rolbyte	pa, crc1, #1
	if_nz	rolbyte	pa, crc0, #1
	if_nz	mergeb	pa    // 4-byte order swapped to bit order of 8 nibbles
	if_nz	xzero	r_crc-0, pa    // 8 nibbles, transmit low nibble first
	if_nz	xzero	r_crc-0, endbit
	if_nz	waitse1

// tx complete, now rx "CRC status"
		dirl	r_pdat-0    // release DAT pins ready for CRC status from SD card
		and	r_pdat-0, #0x3f    // strip away ADDPINS, leaving DAT0

		rep	@.rend2, #6
		wypin	#1, r_clk-0    // one clock pulse
		testp	r_pdat-0   wc    // sample DAT0 before the clock pin transitions
		rcl	timeout, #1    // collect status bits
		waitse1      // wait for each clock done
.rend2
		ret

clocks		long 1 + 512 * 2    // start-bit + nibble count
lnco		long 0x8000_0000UL    // tells RDFAST/WRFAST not to wait for FIFO ready
endbit		long 0xffff_ffffUL
poly		long 0x8408    // CRCNIB shifts right, therefore ordered from left: x0 + x5 + x12 + x16
ptr		long 0x1f6<<19 | COGCRCBUFF<<9    // register PA for ALTI result substitution, and CRC buffer address
crc3		long 0
crc2		long 0
crc1		long 0
crc0		long 0
    }
    __asm {
		wypin	#500, r_clk-0    // trailing clocks
		zerox	timeout, #5    // remove excess bits from "CRC status"
    }

#ifdef SD_DEBUG
    if( timeout == 0x37 )
        __builtin_printf(" BlockWriteError: No CRC response!\n");
    if( timeout == 0b10_101_1 )
        __builtin_printf(" BlockWriteError: CRC mismatch response!\n");
    if( timeout == 0 )
        __builtin_printf(" BlockWriteError: Data busy timeout!\n");
#endif

    return timeout;    // $3F = no response, $00 = busy timeout, %10_101_1 = CRC mismatch, %10_010_1 = SD card happy
}



//-----------------------------------------------------------------------
// Rx data block's DMA parameter setup
//-----------------------------------------------------------------------

static struct rxblk_parms_t {
    uint32_t  p_clk, p_dat, m_leadin, m_nco, m_se2, m_dat, m_clkdiv, m_clks;
} rxblkset;

enum {
      r_clk = SET_PARAMS, r_pdat, r_leadin, r_nco, r_se2, r_mdat, r_clkdiv, r_clks
};


static void  set_rxblkset(
    uint32_t clkdiv,
    uint32_t rxlag )
{
    unsigned  PIN_DAT0 = dat0pin;

    if( rxlag & 1 ) {    // pins registered
        rxlag >>= 1;
        rxlag++;
    } else {    // pins unregistered
        rxlag >>= 1;
    }

    rxblkset.p_clk = clkpin;
    rxblkset.p_dat = PIN_DAT0 | 3<<6;
    rxblkset.m_se2 = 0b100<<6 | PIN_DAT0+3;    // trigger on low level, preferred to falling edge trigger because
        // DAT pins idle high during command and response and also ensures best chance of seeing an early start-bit
    if( clkdiv < 20 )  {
        rxblkset.m_leadin = X_IMM_32X1_1DAC1 | 8 + clkdiv + rxlag;    // skip past the start(S)-bit
        rxblkset.m_clks = 10 | 5<<16;    // start-bit search clock divider
    } else {
        rxblkset.m_leadin = X_IMM_32X1_1DAC1 | 8 + clkdiv*2 + rxlag;    // skip past the start(S)-bit
        rxblkset.m_clks = clkdiv | clkdiv/2<<16;    // start-bit search clock divider
    }
    rxblkset.m_clkdiv = clkdiv | clkdiv/2<<16;
    rxblkset.m_nco = 0x8000_0000UL / clkdiv + (0x8000_0000UL % clkdiv > 0 ? 1 : 0);  // data transfer rate, round up upon a non-zero remainder
#ifdef SD_BLOCK_CRC
    rxblkset.m_dat = X_4P_1DAC4_WFBYTE | PIN_DAT0<<17 | X_PINS_ON | X_ALT_ON | 512 * 2 + 16;  // mode and nibble count
#else
    rxblkset.m_dat = X_4P_1DAC4_WFBYTE | PIN_DAT0<<17 | X_PINS_ON | X_ALT_ON | 512 * 2;  // mode and nibble count
#endif
}


//-----------------------------------------------------------------------
// Receive data blocks from the card
//-----------------------------------------------------------------------

static int  rx_datablocks(    // Remaining block count, 0 = Success
    uint8_t *buf,    // Buffer address to store the received data blocks
    uint32_t blocks,    // Number of data blocks to be read
    uint32_t timeout,    // 100 ms timeout for each start-bit latency, in sysclock ticks
    uint8_t *crcbuf )    // extra hubRAM for holding the final 16 CRC nibbles
{
    uint32_t  crc3, crc2, crc1, crc0;

    __asm const {    // "const" enforces XIP, "volatile" enforces Fcache
		loc	pa, #rxblkset    // fast copy from struct
		setq	#sizeof(rxblkset) / 4 - 1
		rdlong	r_clk-0, pa    // fast copy to registers PR8..PR15

		fltl	r_pdat-0
		setse2	r_se2-0    // trigger on low level, preferred to falling edge trigger
		// because DAT pins idle high during command and response and also ensures best chance
		// of seeing an early start-bit
		pollse1
    }

    __asm volatile {    // "const" enforces XIP, "volatile" enforces Fcache
#ifdef SD_BLOCK_CRC
		modz	_clr   wz    // clear Z flag to create a one-shot for first block
nextblk
	if_z	setq	#512/4+16/8-1    // one SD block + CRC
	if_z	rdlong	COGCRCBUFF, buf    // fast copy to cogRAM before engaging the FIFO
	if_z	add	buf, blkoffset
		wrfast	lnco, buf    // setup FIFO for streamer use
// no room in the buffer for final CRC-16 nibbles, to be handled later
		cmp	blocks, #2   wc    // last block?
	if_c	sub	clocks, #16+2    // stop the clocks before the CRC-16 nibbles
	if_c	sub	r_mdat-0, #16    // truncate DMA to end of buffer
#else
		modz	_clr   wz    // clear Z flag to create a one-shot for first block
		wrfast	lnco, buf    // setup FIFO for streamer use
nextblk
#endif
// locate data block start-bit
// NOTE: Search code has a one SD clock lag.  The start(S)-bit has been and gone by the time it triggers.
//   The following streamer setup begins its sampling ahead of the second data bit arriving.
		wxpin	r_clks-0, r_clk-0    // clock divider for start-bit search
		wypin	lnco, r_clk-0    // start the search
		getct	pb    // timeout begins
		pollse2       // clear stale event
		setxfrq	lnco        // set streamer to sysclock/1 for lead-in timing
		add	pb, timeout    // SDHC 100 ms timeout of start-bit search, Nac (SD spec 4.12.4 note 1)
		setq	pb    // apply the 100 ms timeout to WAITSEn
		waitse2   wc    // wait for DAT3 falling edge, a timeout sets C
		dirl	r_clk-0    // halt clock gen ASAP
		wxpin	r_clkdiv-0, r_clk-0    // clock divider for data block

// start-bit found, now read the data block
	if_nc	xinit	r_leadin-0, #0    // lead-in delay from here at sysclock/1
	if_nc	setq	r_nco-0        // streamer transfer rate (takes effect with buffered command below)
	if_nc	xzero	r_mdat-0, #0     // rx buffered-op, aligned with clock via lead-in
		dirh	r_clk-0    // clock timing starts here
	if_nc	wypin	clocks, r_clk-0    // first pulse outputs during second clock period

#ifdef SD_BLOCK_CRC
	if_nc	call	#crc_check
// prep for next block+CRC
	if_nc	waitse1    // wait for clocking to complete
if_nc_and_z	djnz	blocks, #nextblk

// now collect final block CRC-16 nibbles
if_nc_and_z	wrfast	lnco, crcbuf    // setup FIFO for streamer use, assumes mdat holds stale streamer mode
if_nc_and_z	setxfrq	lnco    // set streamer to sysclock/1 for lead-in timing
if_nc_and_z	dirl	r_clk-0    // reset clock smartpin
if_nc_and_z	setword	r_mdat-0, #16, #0    // reuse streamer mode for collecting the CRC-16 nibbles

if_nc_and_z	xinit	r_leadin-0, #0    // lead-in delay from here at sysclock/1
if_nc_and_z	setq	r_nco-0        // streamer transfer rate (takes effect with buffered command below)
if_nc_and_z	xzero	r_mdat-0, #0     // rx buffered-op, aligned with clock via lead-in
if_nc_and_z	dirh	r_clk-0    // clock timing starts here
if_nc_and_z	wypin	#16+2, r_clk-0    // CRC-16 nibbles + end bit

#ifdef SD_DEBUG
		mov	timeout, #0
	if_c	mov	timeout, #3    // timed out
#endif
if_c_or_nz	ret    // exit on error (timed out or CRC fail)
		waitse1    // wait for clocking to complete

// copy block and CRC to cogRAM
		setq	#512/4-1    // 512 bytes, one SD block, 4 bytes per longword
		rdlong	COGCRCBUFF, buf    // fast copy to cogRAM
		setq	#16/8-1    // 16 CRC-16 nibbles, 8 nibbles per longword
		rdlong	COGCRCBUFF+512/4, crcbuf
// do the check
		call	#crc_check
	if_nz	add	blocks, #1    // indicate a failed final CRC
#else
	if_nc	waitse1    // wait for clocking to complete
	if_nc	djnz	blocks, #nextblk
#ifdef SD_DEBUG
		mov	timeout, #0
	if_c	mov	timeout, #3    // timed out
#endif
#endif
		ret

lnco		long 0x8000_0000UL    // lead-in rate, sysclock/1
clocks		long 512 * 2 + 16 + 2    // clock pulses, nibble count + 16-bit CRC + end-bit
#ifdef SD_BLOCK_CRC
poly		long 0x8408    // CRCNIB shifts right, therefore ordered from left: x0 + x5 + x12 + x16
cogaddr		long 0x1f6<<19 | COGCRCBUFF<<9    // register PA for ALTI result substitution, and CRC buffer address
blkoffset	long 512

crc_check
		mov	crc3, #0
		mov	crc2, #0
		mov	crc1, #0
		mov	crc0, #0
		mov	pb, cogaddr
// check CRC of prior read SD block
	if_z	rep	@.rend, #512/4+16/8    // one SD block + CRC
	if_z	alti	pb, #0b100_111_000    // next D-field substitution, then increment PB, result goes to PA
	if_z	movbyts	pa, #0b00_01_10_11    // byte swap within longword
	if_z	splitb	pa    // 8-nibble order swapped to bit order of 4 bytes
	if_z	setq	pa
	if_z	crcnib	crc3, poly
	if_z	crcnib	crc3, poly
	if_z	crcnib	crc2, poly
	if_z	crcnib	crc2, poly
	if_z	crcnib	crc1, poly
	if_z	crcnib	crc1, poly
	if_z	crcnib	crc0, poly
	if_z	crcnib	crc0, poly
.rend
// pass/fail
		or	crc3, crc2
		or	crc1, crc0
		or	crc3, crc1   wz    // Z set for pass, clear for fail
		long 0xfd64002d    // RET
#endif
    }

#ifdef SD_DEBUG
    if( blocks ) {
        __builtin_printf(" BlockReadError:  blocks=%d  err=%d  crc=%08x  ", blocks, timeout, crc3);
        if( timeout == 3 )
            __builtin_printf("start-bit timed out!\n");
        else
            __builtin_printf("CRC failed!\n");
    }
#endif

    return blocks;    // number of blocks remaining, 0 = Success
}



//-----------------------------------------------------------------------
// Command response's DMA parameter setup
//-----------------------------------------------------------------------

static struct cmdresp_parms_t {
    uint32_t  p_clk, p_cmd, m_leadin, m_nco, timeout, m_resp, m_clkr, m_clks;
} crset;

enum {
      r_clk = SET_PARAMS, r_cmd, r_leadin, r_nco, r_timeout, r_resp, r_clkr, r_clks
};


static void  set_respset(
    uint32_t clkdiv,
    uint32_t rxlag )
{
    unsigned  PIN_CMD = cmdpin;

    if( rxlag & 1 ) {    // pins registered
        rxlag >>= 1;
        rxlag++;
    } else {    // pins unregistered
        rxlag >>= 1;
    }

    crset.p_clk = clkpin;
    crset.p_cmd = PIN_CMD;
    crset.m_resp = X_1P_1DAC1_WFBYTE | PIN_CMD<<17 | X_PINS_ON | X_ALT_ON;    // + bit count
    crset.m_nco = 0x8000_0000UL / clkdiv + (0x8000_0000UL % clkdiv > 0 ? 1 : 0);  // data transfer rate, round up upon a non-zero remainder
    if( clkdiv < 20 )  {    // start-bit search ends on T-bit instead of S-bit
        crset.m_leadin = X_IMM_32X1_1DAC1 | 8 + rxlag;    // store start(S)-bit ahead of clock pulse
        crset.timeout = 10 * 64;    // max of 64 clocks for response to start, Ncr (SD spec 4.12.4)
        crset.m_clks = 10 | 5<<16;    // start-bit search clock divider
    } else {
        crset.m_leadin = X_IMM_32X1_1DAC1 | 8 + clkdiv + rxlag;    // store start(S)-bit ahead of clock pulse
        crset.timeout = clkdiv * 64;    // max of 64 clocks for response to start, Ncr (SD spec 4.12.4)
        crset.m_clks = clkdiv | clkdiv/2<<16;    // start-bit search clock divider
    }
    crset.m_clkr = clkdiv | clkdiv/2<<16;    // response clock divider
}


//-----------------------------------------------------------------------
// Receive the card's command-response packet
//-----------------------------------------------------------------------

static int  rx_response(    // 0:Start-bit timed out
    uint8_t *resp,    // Buffer address for response bytes
    uint32_t len,    // Number of bytes in response, including the start-end framing bits
    uint32_t clocks )    // Additional clocks for card's post-response idle transition
{
    uint32_t lnco = 0x8000_0000UL;

    __asm const {    // "const" enforces XIP, "volatile" enforces Fcache
		loc	pb, #crset    // fast copy from struct
		setq	#sizeof(crset) / 4 - 1
		rdlong	r_clk-0, pb    // fast copy to registers PR8..PR15

		shl	len, #3
		setword	r_resp-0, len, #0    // add bit count to DMA mode word, crset.m_resp
		add	clocks, len
		mov	pb, #0b010<<6    // trigger on falling edge of CMD pin - start-bit arrived
		or	pb, r_cmd-0
//		setse2	#0    // cancel triggering before reuse - not needed if POLLSE2 is used
		setse2	pb
    }

    __asm volatile {    // "const" enforces XIP, "volatile" enforces Fcache
		waitse1    // wait for end of command tx clocks

// locate response start-bit
// NOTE: Search code has a one SD clock lag.  The start(S)-bit has been and gone by the time it triggers.
//   The following streamer setup samples the T bit twice before further data bits arrive.
		outh	r_cmd-0    // when driven, momentarily force CMD pin high before start-bit search
		wxpin	r_clks-0, r_clk-0    // clock divider for start-bit search
		wypin	lnco, r_clk-0    // start the search
		fltl	r_cmd-0    // pin drive release before SD card takes over
		pollse2       // clear stale event
		wrfast	lnco, resp    // engage the FIFO
		setxfrq	lnco    // set sysclock/1 for lead-in timing
		getct	pb    // timeout begins
		add	pb, r_timeout-0    // max of 64 clocks for response to start, Ncr (SD spec 4.12.4)
		setq	pb    // apply the 64 clock timeout to WAITSEn
		waitse2   wc    // wait for CMD start bit, C set if timed-out
		dirl	r_clk-0    // halt clock gen ASAP
		wxpin	r_clkr-0, r_clk-0    // revert to regular clock gen

// start-bit found, now read the response (NOTE: start-bit is part of first byte)
	if_nc	xinit	r_leadin-0, #0    // lead-in delay from here at sysclock/1
	if_nc	setq	r_nco-0    // streamer transfer rate (takes effect with buffered command below)
	if_nc	xzero	r_resp-0, #0     // rx buffered-op aligned with clock via lead-in
		dirh	r_clk-0    // clock timing starts here
	if_nc	wypin	clocks, r_clk-0    // first clock pulse outputs during second clock period

		// twiddle thumbs
	if_c	mov	len, #0    // start-bit timed out

	if_nc	waitxfi      // wait for rx data completion before resuming hubexec
    }
#ifdef SD_DEBUG
    if( !len )
        __builtin_printf(" Response start-bit timeout! ");
#endif

    return len;    // 0:Start-bit timed out
}



//-----------------------------------------------------------------------
// Command's DMA parameter setup
//-----------------------------------------------------------------------

static struct cmd_parms_t {
    uint32_t  p_clk, p_cmd, m_leadin, m_nco, m_ca;
} cmdset;

enum {
      r_clk = SET_PARAMS, r_cmd, r_leadin, r_nco, r_ca
};


static void  set_cmdset(
    uint32_t clkdiv,
    uint32_t CMDAT_REG )
{
    unsigned  PIN_CMD = cmdpin;

    CMDAT_REG = CMDAT_REG & 1;    // extract lsb from rxlag

    cmdset.p_clk = clkpin;
    cmdset.p_cmd = PIN_CMD;
    cmdset.m_leadin = X_IMM_32X1_1DAC1 | 5 + CLK_REG - CMDAT_REG + (CLK_POL ? clkdiv : clkdiv/2);
    cmdset.m_nco = 0x8000_0000UL / clkdiv + (0x8000_0000UL % clkdiv > 0 ? 1 : 0);  // data transfer rate, round up upon a non-zero remainder
    cmdset.m_ca = X_IMM_32X1_1DAC1 | PIN_CMD<<17 | X_PINS_ON | 32;    // mode and bit count
}


//-----------------------------------------------------------------------
// Transmit a command packet to the card
//-----------------------------------------------------------------------

static int  tx_command(    // 0:CMD pin stuck low
    uint32_t cmd,    // Command
    uint32_t arg )    // Argument
{
    __asm const {    // "const" enforces XIP, "volatile" enforces Fcache
		loc	pb, #cmdset    // fast copy from struct
		setq	#sizeof(cmdset) / 4 - 1
		rdlong	r_clk-0, pb    // fast copy to cogRAM

		setse1	#0    // cancel triggering before reuse - not needed if POLLSE1 is used
		mov	pb, r_clk-0
		or	pb, #0b001<<6    // trigger on rising edge - clocks completed
		wypin	#1, r_clk-0    // ensure card releases the CMD bus
		setse1	pb    // must be immediately after WYPIN to catch IN rising

		or	cmd, #0x40    // add T bit in front of command
		shl	cmd, #24    // command at [31:24] is needed for later CRC
		mov	pb, arg
		shr	pb, #8
		or	pb, cmd
		rev	pb    // first 32 bits prep'd for streamer
		setxfrq	##0x8000_0000UL    // set sysclock/1 for lead-in timing
		waitse1        // wait for clock pulse to complete
		fltl	r_clk-0    // reset clock gen smartpin, needed to instruction align the samrtpin cycle
		testp	r_cmd-0   wc    // send new command only if CMD pin has returned high
// C is set when CMD pin is high
	if_c	drvl	r_cmd-0    // drive CMD pin low, start-bit, ready for streamer output

// begin command phase
	if_c	xinit	r_leadin-0, #0    // lead-in delay at sysclock/1
	if_c	setq	r_nco-0    // data transfer rate, takes effect on XZERO below (buffered command)
	if_c	xzero	r_ca-0, pb    // place first 32 bits in tx buffer, aligned with clock via lead-in delay
		dirh	r_clk-0    // clock timing starts here, first clock pulse occurs in smartpin's second period
		wypin	#6*8+1, r_clk-0    // SD clock pulses, 6 bytes + 1 pulse to kick off a response

// CRC7 algorithm, SD spec 4.5
	if_c	setq	cmd    // first byte, CMD<<24
	if_c	mov	pb, #0
	if_c	crcnib	pb, #0x48    // CCITT polynomial is x7 + x3 + 1
	if_c	crcnib	pb, #0x48    // 0x09 reversed and shifted for CRCNIB
	if_c	setq	arg     // + four bytes
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48

	if_c	or	pb, #0x180    // add end-bit after the CRC
	if_c	rev	arg
	if_c	rolbyte	pb, arg, #3    // insert last byte of arg ahead of CRC

	if_c	xzero	r_ca-0, pb     // place remaining 16 bits in buffer for when first 32 bits completes
	if_nc	mov	cmd, #0    // CMD pin stuck low - Probably needs power down
    }
#ifdef SD_DEBUG
    if( !cmd )
        __builtin_printf(" CMD pin stuck low! ");
#endif

    return cmd;    // 0:CMD pin stuck low
}



//-----------------------------------------------------------------------
// Setup the SD clock pin. Also configure DMA parameter sets
//-----------------------------------------------------------------------

static void  sd_clockconf(
    uint32_t clkdiv,
    uint32_t rxlag )
{
    uint32_t  pinmode;

    // update the setsets
    set_cmdset(clkdiv, rxlag);
    set_respset(clkdiv, rxlag);
    set_rxblkset(clkdiv, rxlag);
    set_txblkset(clkdiv, rxlag);

    // configure pins
    pinmode = (rxlag&1 ? P_SYNC_IO : 0) | (RX_SCHMITT ? P_SCHMITT_A : 0);
    _wrpin(cmdpin, pinmode);
    _wrpin(dat0pin | 3<<6, pinmode);

    // SD clock gen smartpin
    pinmode = P_PULSE | P_OE | (CLK_POL ? P_INVERT_OUTPUT : 0) | P_SCHMITT_A | (CLK_REG ? P_SYNC_IO : 0);
    _pinstart(clkpin, pinmode, clkdiv | clkdiv/2<<16, 0);
}



//-----------------------------------------------------------------------
// Send a common housekeeping command packet to the card
//-----------------------------------------------------------------------

static int  send_cmd(    // 0:Fail, 1:Success
    uint32_t cmd,    // Command
    uint32_t arg,    // Argument
    uint8_t *resp )    // Buffer address for response bytes
{
#ifdef SD_DEBUG
//    __builtin_printf(" CMD%d - ", cmd);
#endif
    tx_command(cmd, arg);
    if( !resp ) {    // expecting no response
        unsigned  PIN_CLK = clkpin;
        unsigned  PIN_CMD = cmdpin;
        __asm {
		waitse1    // wait for end of command tx clocks
		outh	PIN_CMD    // momentarily force CMD pin high before new clocks
		wypin	#8, PIN_CLK    // trailing clocks
		fltl	PIN_CMD    // pin drive release before SD card takes over
		waitse1       // wait for end of command tx clocks
		wypin	#500, PIN_CLK    // trailing clocks
        }
        return 1;
    }
    if( rx_response(resp, 6, -500) )    // R1,R1b,R3,R6,R7 responses
        if( resp[5] == crc7sd(resp, 5) )
            return 1;
#ifdef SD_DEBUG
    __builtin_printf(" CMD%d error! ", cmd);
#endif
    return 0;
}



//-----------------------------------------------------------------------
// Send a R2 housekeeping command packet to the card
//-----------------------------------------------------------------------

static int  send_cmd_r2(    // 0:Fail, 1:Success
    uint32_t cmd,    // Command
    uint32_t arg,    // Argument
    uint8_t *resp )    // Buffer address for response bytes
{
#ifdef SD_DEBUG
//    __builtin_printf(" CMD%d - ", cmd);
#endif
    tx_command(cmd, arg);
    if( rx_response(resp, 17, -500) )    // R2 response, response length, trailing clocks
        if( resp[16] == crc7sd(&resp[1], 15) )
            return 1;
#ifdef SD_DEBUG
    __builtin_printf(" CMD%d error! ", cmd);
#endif
    return 0;
}



//-----------------------------------------------------------------------
// Send an application specific command packet to the card
//-----------------------------------------------------------------------

static int  send_acmd(    // 0:Fail, 1:Success
    uint32_t acmd,    // App command
    uint32_t arg,    // Argument
    uint8_t *resp )    // Buffer address for response bytes
{
    if( send_cmd(55, rca16<<16, resp) ) {    // ACMD prefix
#ifdef SD_DEBUG
//    __builtin_printf(" ACMD%d - ", acmd);
#endif
        tx_command(acmd, arg);
        if( rx_response(resp, 6, -500) )
            return 1;    // skip CRC processing since not all ACMDs have one
    }
#ifdef SD_DEBUG
    __builtin_printf(" ACMD%d error! ", acmd);
#endif
    return 0;
}



//-----------------------------------------------------------------------
// Wait for card ready - Checks the DAT0 pin write-data-busy indicator
//-----------------------------------------------------------------------

static int  wait_ready(    // 0:Card Busy, 1:Card Ready
    uint32_t timeout )    // 100/250/500 ms interval in sysclock ticks
{
    uint32_t  m_se2;
    int  ready;

    timeout += _cnt();
    _wypin(clkpin, -1);
    m_se2 = 0b110<<6 | dat0pin;    // trigger on high level - card ready
    ready = 0;   // Busy

    __asm {    // "const" enforces XIP, "volatile" enforces Fcache
		setse2	#0    // cancel triggering before reuse - not needed if POLLSE2 is used
		setse2	m_se2
		setq	timeout
		waitse2   wc
	if_nc	mov	ready, #1    // Not busy
    }

#ifdef SD_DEBUG
    if( !ready )
        __builtin_printf(" Busy timeout! ");
#endif

    return ready;
}



//-----------------------------------------------------------------------
//-----------------------------------------------------------------------

static int32_t  calibrate_rxlag(    // 0:Failed calibration,  1..16:New rxlag
    uint8_t *resp )    // buffer address for response bytes
{
    uint32_t  CLK_DIV = clkdivider;
    uint32_t  rca = rca16 << 16;    // stored copy of the card's RCA register
    int32_t  rxlag, repeats, highest = 0, lowest = 0;

    send_cmd(7, 0, NULL);    // CMD7  DESELECT_CARD - "tran" to "standby" state, no response

    for( rxlag = 1; rxlag <= 30; rxlag++ )
    {
        sd_clockconf(CLK_DIV, rxlag);

        repeats = 80;
        do {
#ifdef SD_DEBUG
            putchar('.');
#endif
            if( !send_cmd_r2(10, rca, resp) )    // CMD10  SEND_CID - R2 response
                break;
            if( memcmp(cid, resp, 17) )    // compare stored copy of CID register
                break;
        } while( --repeats );

        if( !repeats ) {    // pass
            highest = rxlag;
            if( !lowest )
                lowest = rxlag;
        } else
            if( highest )    // done
                break;
#ifdef SD_DEBUG
        puts("");
#endif
    }

    if( highest )  {
        rxlag = lowest + (highest - lowest + 1) / 2;
//        rxlag = rxlag > lowest ? rxlag : highest;
        rxlagcomp = rxlag;

        sd_clockconf(CLK_DIV, rxlag);
    } else
        rxlag = 0;
#ifdef SD_DEBUG
    __builtin_printf( "  rxlag=%d selected  Lowest=%d Highest=%d\n", rxlag, lowest, highest );
#endif
    send_cmd(7, rca, resp);    // CMD7  SELECT_CARD - "standby" to "tran" state, R1b response

    return rxlag;
}



//-----------------------------------------------------------------------
//-----------------------------------------------------------------------

static int  calibrate_rxlag_data(    // 0:Failed calibration,  1..16:New rxlag
    uint8_t *buf,     // address of stored copy of the card's CID register
    uint8_t *resp )    // buffer address for response bytes
{
    uint32_t  timeout = _clockfreq() / 4;

    if( !wait_ready(timeout) )
        goto errorlabel;

    tx_command(19, 0);    // CMD19  SEND_TUNING
    if( !rx_response(resp, 6, 2) )    // R1 response
        goto errorlabel;
    rx_datablocks(buf, 1, timeout, resp);    // collect the data block

    for( int i = 0; i <= 139; i++ )
        __builtin_printf(" %02x", buf[i]);
    __builtin_printf("\n");

    return 1;

errorlabel:
#ifdef SD_DEBUG
    __builtin_printf(" SEND_TUNING error!\n");
#endif
    return 0;
}



//-----------------------------------------------------------------------
//-----------------------------------------------------------------------

static LBA_t  disc_size(
    uint8_t *csd )
{
    uint32_t  cs = __builtin_bswap32(*(uint32_t *)&csd[6]);
    int  n;

    if( csd[0]>>6 ) {    // SDC ver 2.00
        cs = (cs & 0xfff_ffff) + 1;  // 0.5 MiByte granularity
        n = 10;
    } else {    // SDC ver 1.00
        cs = (cs>>14 & 0xfff) + 1;    // C_SIZE
        n = (__builtin_bswap16(*(uint16_t *)&csd[9])>>7 & 0x7) + 2    // C_SIZE_MULT
            + (csd[5] & 15) - 9;    // READ_BL_LEN
    }

    return (LBA_t)cs << n;    // 32/64-bit block count
}



//--------------------------------------------------------------------------
//
//   Public Functions
//
//--------------------------------------------------------------------------


//-----------------------------------------------------------------------
// Get Disk Status                                                       
//-----------------------------------------------------------------------

DSTATUS disk_status(
    BYTE drv )       // Drive number (always 0)
{
    if( drv )
        return STA_NOINIT;

    return Stat;
}



//-----------------------------------------------------------------------
// Initialize Disk Drive
//-----------------------------------------------------------------------
enum {
    CMD8CHECKPAT = 0x100|0x5a  // 3V3 bit and check pattern, SD spec 4.3.13
};


DSTATUS disk_initialize (
    BYTE drv )        // Physical drive nmuber (0)
{
    unsigned  PIN_CLK = clkpin;
    uint32_t  queuedepth, rca, tmr, a41arg, CLK_DIV = clkdivider;
    uint8_t  *resp = __builtin_alloca(20);    // response buffer, in hubRAM
    DSTATUS  status;

    status = STA_NOINIT | STA_NODISK;
    Stat = status;

    if( drv ) {
#ifdef SD_DEBUG
        __builtin_printf("bad drv %d\n", drv);
#endif
        return status;
    }
//_pinl(56);    // diag
    if( !sdcard_power() )    // attempt to power cycle the SD card, ignore outcome
        return status;

// setup SD clock smartpin, and other pins, and fill out initial setsets
    tmr = (_clockfreq() + 200_000UL) / 400_000UL;    // divider for 400 kHz SD clock
    rxlagcomp = 0;
    sd_clockconf(tmr, 0);
#ifdef SD_DEBUG
    __builtin_printf(" SD clock divider set to sysclock/%d (400 kHz)\n", tmr);
#endif

// give it a kick
    _wypin(PIN_CLK, 500);    // SD spec 6.4.1
    _waitms(1);
    send_cmd(0, 0, NULL);    // CMD0  GO_IDLE_STATE - in case the power cycle didn't happen
    send_cmd(8, CMD8CHECKPAT, resp);    // CMD8  SEND_IF_COND - 3.3V and check pattern, SD spec 4.3.13, R7 response
    _waitms(1);

// start conversing with SD card
    send_cmd(0, 0, NULL);    // CMD0  GO_IDLE_STATE - no response expected
//    if( !send_cmd(8, CMD8CHECKPAT, resp) )    // CMD8  SEND_IF_COND - 3.3V and check pattern, SD spec 4.3.13, R7 response
//        goto faillabel;    // no response, probably card missing
    if( send_cmd(8, CMD8CHECKPAT, resp) ) {    // CMD8  SEND_IF_COND - 3.3V and check pattern, SD spec 4.3.13, R7 response
        if( __builtin_bswap32(*(uint32_t *)&resp[1]) != CMD8CHECKPAT )// confirms SD v2.0+, with valid voltage and check pattern
            goto faillabel;
        a41arg = 1<<30 | 1<<20 | 1<<28;    // set host as HCS capable, 3.3 Volt, and request XPC (full power)
    } else
        a41arg = 1<<20;    // card is not HCS capable, SD v1.0, 3.3 Volt

#ifdef SD_DEBUG
    __builtin_printf(" Card idle OK\n");
#endif
    status = STA_NOINIT;

// find card type
    rca = 0;
    rca16 = 0;    // "idle" state uses RCA=0 in ACMDs
    tmr = _getms();
    do {
        if( !send_acmd(41, a41arg, resp) )   // ACMD41  SD_SEND_OP_COND - "idle" to "ready" state, R3 response
            goto faillabel;
        rca = __builtin_bswap32(*(uint32_t *)&resp[1]);
//            rca = resp[4];
//            if( rca>>7 )    // valid Ready bit, card has switched from "idle" to "ready" state
        if( rca>>31 )    // valid Ready bit, card has switched from "idle" to "ready" state
            break;
        _wypin(PIN_CLK, 400);
        _waitms(1);    // card busy, polling faster than 50 ms, SD spec 4.4
    } while( _getms() - tmr < 1000 );    // SD spec 4.2.3, 1.0 sec timeout, probably gone "inactive"
#ifdef SD_DEBUG
    __builtin_printf(" OCR register %08x  - ",rca);
#endif
    rca >>= 30;
    if( rca != 3 )    // check CCS bit
        ledpin |= 1;    // byte addressed card type
#ifdef SD_DEBUG
    if( rca == 3 )
        __builtin_printf("SDHC/SDXC Card\n");    // block addressed card type
    else
        __builtin_printf("SDSC Card\n");    // byte addressed card type
#endif

//-----------------------------------------
// CMD11 voltage switch not used
//   Default 3.3V is ideal for Prop2 host
//   UHS features consequently ignored
//-----------------------------------------

    // CMD2 then CMD3 are required back-to-back.  If the sequence is
    //   unsuccesful then CMD3 will timeout and return a zero value
    send_cmd_r2(2, 0, resp);    // CMD2  ALL_SEND_CID - "ready" to "ident" state, R2 response (CID)
    // also next R1/R6 card status (CMD3) will indicate a partial ACMD state
    if( !send_cmd(3, 0, resp) )    // CMD3  publish new RCA - "ident" to "standby" state, R6 response
        goto faillabel;

    rca = __builtin_bswap16(*(uint16_t *)&resp[1]);
    rca16 = rca;
    rca = rca << 16;
#ifdef SD_DEBUG
    __builtin_printf(" Data Transfer Mode entered - Published RCA %08x\n", rca);
#endif

    send_cmd_r2(10, rca, resp);    // CMD10  SEND_CID, R2 response
    memcpy(cid, resp, 17);
#ifdef SD_DEBUG
    __builtin_printf(" CID register backed up\n");
#endif

    if( !send_cmd(7, rca, resp) )    // CMD7  SELECT_CARD - "standby" to "tran" state, R1b response
        goto faillabel;

    status = 0;   // card is good to go

    // Need 4-bit bus width for data blocks
    if( !send_acmd(6, 2, resp) )  // ACMD6  SET_BUS_WIDTH - Engage 4-bit parallel data, R1 response
        goto faillabel;
#ifdef SD_DEBUG
    __builtin_printf(" 4-bit data interface engaged\n");
#endif

    uint8_t  *buff = __builtin_alloca(512);    // data buffer, in hubRAM
    uint32_t  timeout = _clockfreq()/4;    // 250 ms

    // Support for caching is not yet implemented, but sounds a promising approach
    send_acmd(13, 0, resp);    // ACMD13  SEND_SD_STATUS (SSR), spec 4.10.2
    rx_datablocks(buff, 1, timeout, resp);    // data length is 64 bytes, CRC will fail
#ifdef SD_DEBUG
//    __builtin_printf("ACMD13: ");
//    for( tmr = 0; tmr <= 63; tmr++ )
//        __builtin_printf(" %02x", buff[tmr]);
//    __builtin_printf("\n");
    tmr = buff[8] * 2;
    if( tmr > 8 )
        tmr = -1;
    if( tmr == 8 )
        tmr = 10;
    __builtin_printf(" Speed Class = C%d", tmr);    // spec 4.10.2.2
    __builtin_printf("  UHS Grade = U%d", buff[14] >> 4);    // spec 4.10.2.8
    __builtin_printf("  Video Class = V%d", buff[15]);    // spec 4.10.2.10
    __builtin_printf("  App Class = A%d\n", buff[21] & 0xf);    // spec 4.10.2.13
#endif

    if( !CLK_POL ) {
        // CMD6  SWITCH_FUNC (Spec 4.3.10 and 4.10.2):  Select High-Speed Access Mode
        //   This primarily doubles the card's allotted power limit to 200 mA
        //   It also phase shifts the clock-data relationship
        send_cmd(6, 0x80_fffff1, resp);    // CMD6  SWITCH_FUNC, Set G1=1 (High Speed)
        // Oddly, CMD6 reads the function table as a data block
        rx_datablocks(buff, 1, timeout, resp);    // func table data length is 64 bytes, CRC will fail
//    _wypin(PIN_CLK, 20);    // SD spec 4.3.10.1, fig 4-14
    }
    // Now to verify the function switch
    send_cmd(7, 0, NULL);    // CMD7  DESELECT_CARD - "tran" to "standby" state, no response
    if( !send_cmd_r2(9, rca, resp) )    // CMD9  SEND_CSD - R2 response
        goto faillabel;
    if( resp[4] != 0x5a ) {    // TRAN_SPEED - SD spec 5.3.3, tables 5-16 and 5-6
#ifdef SD_DEBUG
        __builtin_printf(" Default-Speed access mode\n");
#endif
        if( !CLK_POL )
            goto faillabel;
    } else {
#ifdef SD_DEBUG    // 32h=25MHz, 5Ah=50MHz, 0Bh=100MHz, 2Bh=200MHz
        __builtin_printf(" High-Speed access mode engaged\n");
#endif
    }
#ifdef SD_DEBUG
    tmr = disc_size(&resp[1]);    // block count
    __builtin_printf(" Card User Capacity = %d MiB\n", tmr / 2048);
#endif

    // Time for full speed clocking - After optional High-Speed access mode has been set!
    // Important for timing to stay loose until now because the calibration process
    // phase aligns the rx clock-data relationship - which itself changes between
    // the speed modes.  Calibration would mess up before this point.
//    sd_clockconf(CLK_DIV, 0);    // set full speed SD clock divider
#ifdef SD_DEBUG
    tmr = _clockfreq() / (CLK_DIV * 100_000);
    __builtin_printf(" SD clock divider set to sysclock/%d (%d.%d MHz)\n",
                     CLK_DIV, tmr/10, tmr%10);
#endif

    if( !calibrate_rxlag(resp) )    // perform rx calibration, also selects card and sets full speed clock
        goto faillabel;
//    send_cmd(7, rca, resp);    // CMD7  SELECT_CARD - "standby" to "tran" state, R1b response
#ifdef SD_DEBUG
/*    if( send_cmd_r2(10, rca, resp) ) {    // CMD10  SEND_CID, R2 response
        __builtin_printf("CID decode:  ManID=%02X   OEMID=%c%c",resp[1],resp[2],resp[3]);
        __builtin_printf("  Name=%c%c%c%c%c\n",resp[4],resp[5],resp[6],resp[7],resp[8]);
        __builtin_printf("  Ver=%x.%x   Serial=%08X   Date=%d-%d\n", resp[9]>>4, resp[9]&0xf,
                __builtin_bswap32(*(int32_t*)&resp[10]), 2000+((resp[14]<<4|resp[15]>>4)&0xff), resp[15]&0xf );
    }*/
    __builtin_printf(" CID decode:  ManID=%02X   OEMID=%c%c",cid[1],cid[2],cid[3]);
    __builtin_printf("  Name=%c%c%c%c%c\n",cid[4],cid[5],cid[6],cid[7],cid[8]);
    __builtin_printf("  Ver=%x.%x   Serial=%08X   Date=%d-%d\n", cid[9]>>4, cid[9]&0xf,
                __builtin_bswap32(*(int32_t*)&cid[10]), 2000+((cid[14]<<4|cid[15]>>4)&0xff), cid[15]&0xf );

    __builtin_printf("SD Card Init Successful\n");
#endif
    Stat = status;

    return status;


faillabel:
#ifdef SD_DEBUG
    __builtin_printf("SD Init Failure!  :-(\n");
#endif
    _wypin(PIN_CLK, 500);
    _waitus(20);
    status = STA_NOINIT;
    Stat = status;

    return status;
}



//-----------------------------------------------------------------------
// Read Block(s)
//-----------------------------------------------------------------------

DRESULT disk_read(
    BYTE drv,       // Physical drive nmuber (0)
    BYTE *buff,     // Pointer to the data buffer to store read data
    LBA_t firstblk,    // Start block number (LBA)
    LBA_t blocks )     // Block count
{
    uint8_t  *resp = __builtin_alloca(20);    // response buffer, in hubRAM
    uint32_t  timeout = _clockfreq();    // 250 ms
    unsigned  PIN_LED = ledpin;    // Doubles up as SDSC byte addressing flag in bit0

    if( disk_status(drv) & STA_NOINIT )
        return RES_NOTRDY;

#ifdef SD_DEBUG_PERFORMANCE
    if( blocks < 2 )
        __builtin_printf(" RD%x %d ", firstblk, _getus());
    else
        __builtin_printf(" RD%x+%x %d ", firstblk, blocks, _getus());
#endif
    if( PIN_LED & 1 )    // SDSC card type attached to the pin number
        firstblk <<= 9;    // translate block to byte address
    PIN_LED >>= 8;    // just the pin number

    if( !wait_ready(timeout) )    // examine busy on DAT0
        return RES_ERROR;

    if( blocks < 2 ) {
        tx_command(17, firstblk);    // CMD17  READ_SINGLE_BLOCK
        if( !rx_response(resp, 6, 2) )    // R1 response, response length, trailing clocks
            return RES_ERROR;

        _pinl(PIN_LED);    // activity LED on
        blocks = rx_datablocks(buff, 1, timeout, resp);
        _pinh(PIN_LED);    // activity LED off

        if( !blocks )
            return RES_OK;

    } else {
        tx_command(18, firstblk);    // CMD18  READ_MULTIPLE_BLOCK
        if( !rx_response(resp, 6, 2) )    // R1 response, response length, trailing clocks
            return RES_ERROR;

        _pinl(PIN_LED);    // activity LED on
        blocks = rx_datablocks(buff, blocks, timeout, resp);
        _pinh(PIN_LED);    // activity LED off

        tx_command(12, 0);    // CMD12  STOP_TRANSMISSION
        if( rx_response(resp, 6, -500) )    // don't much care if this works
            if( !blocks )
                return RES_OK;
    }

    calibrate_rxlag(resp);

    return RES_ERROR;
}



//-----------------------------------------------------------------------
// Write Block(s)
//-----------------------------------------------------------------------

DRESULT disk_write(
    BYTE drv,            // Physical drive nmuber (0)
    const BYTE *buff,    // Pointer to the data to be written
    LBA_t firstblk,         // Start block number (LBA)
    LBA_t blocks )          // Block count
{
    uint8_t  *resp = __builtin_alloca(20);    // response buffer, in hubRAM
    uint32_t  timeout = _clockfreq()/4UL;    // 250 ms
    unsigned  PIN_LED = ledpin;    // Doubles up as SDSC byte addressing flag in bit0

    if( disk_status(drv) & STA_NOINIT )
        return RES_NOTRDY;

#ifdef SD_DEBUG_PERFORMANCE
    if( blocks < 2 )
        __builtin_printf(" WR%x %d ", firstblk, _getus());
    else
        __builtin_printf(" WR%x+%x %d ", firstblk, blocks, _getus());
#endif
    if( PIN_LED & 1 )    // SDSC card type attached to the pin number
        firstblk <<= 9;    // translate block to byte address
    PIN_LED >>= 8;    // just the pin number

    if( !wait_ready(timeout) )    // examine busy on DAT0
        return RES_ERROR;

    if( blocks < 2 ) {
        tx_command(24, firstblk);    // CMD24  WRITE_BLOCK
        if( !rx_response(resp, 6, 2) )    // R1 response, response length, trailing clocks
            return RES_ERROR;

        _pinl(PIN_LED);    // activity LED on
        blocks = tx_datablock(buff, timeout);
        _pinh(PIN_LED);    // activity LED off
        if( blocks == 0b10_010_1 )
            return RES_OK;

    } else {
#ifdef SD_USE_ACMD23
        if( !send_acmd(23, blocks, resp) )    // ACMD23  SET_WR_BLK_ERASE_COUNT
            return RES_ERROR;
#endif
        tx_command(25, firstblk);    // CMD25  WRITE_MULTIPLE_BLOCK
        if( !rx_response(resp, 6, 2) )    // R1 response, response length, trailing clocks
            return RES_ERROR;

        _pinl(PIN_LED);    // activity LED on
        do {
            if( tx_datablock(buff, timeout) != 0b10_010_1 )
                break;   // other than CRC matched
            buff += 512;
        } while( --blocks );
        _pinh(PIN_LED);    // activity LED off

        tx_command(12, 0);    // CMD12  STOP_TRANSMISSION
        if( rx_response(resp, 6, -500) )    // R1b response, response length, trailing clocks
            if( !blocks )
                return RES_OK;
    }

    calibrate_rxlag(resp);

    return RES_ERROR;

}



//-----------------------------------------------------------------------
// Miscellaneous Functions
//-----------------------------------------------------------------------

DRESULT disk_ioctl(
    BYTE drv,    // Physical drive nmuber (0)
    BYTE ctrl,    // Control code
    void *buff )    // Buffer to send/receive control data
{
    if( disk_status(drv) & STA_NOINIT )    // Check if card is in the socket
        return RES_NOTRDY;

    uint8_t  *resp = __builtin_alloca(20);    // response buffer, in hubRAM
    uint32_t  rca = rca16<<16;
    DRESULT  res = RES_ERROR;

    switch( ctrl ) {
        case CTRL_SYNC :    // Make sure that no pending write process
        {
            uint32_t  timeout = _clockfreq()/4UL;    // 250 ms
#ifdef SD_DEBUG_PERFORMANCE
            __builtin_printf(" SYNC %d ", _getus());
#endif
            if( wait_ready(timeout) )    // selects card to examine busy on DAT0
                res = RES_OK;
            break;
        }

        case GET_SECTOR_COUNT :    // Get number of blocks on the disc
#ifdef SD_DEBUG_PERFORMANCE
            __builtin_printf(" CAPACITY ");
#endif
            if( send_cmd_r2(9, rca, resp) ) {    // SEND_CSD, R2 response
                *(LBA_t *)buff = disc_size(&resp[1]);
                res = RES_OK;
            }
            break;

        case GET_BLOCK_SIZE :    // Get erase block size in unit of sector
            *(LBA_t *)buff = 128;
            res = RES_OK;
            break;

        default:
            res = RES_PARERR;
    }

    return res;
}



DRESULT disk_setpins(
    BYTE drv,
    unsigned CLK_DIV,
    int PIN_CLK,
    int PIN_CMD,
    int PIN_DAT0,
    int PIN_POW,
    int PIN_LED )
{
    if( drv )
        return RES_NOTRDY;

#ifdef SD_DEBUG
    __builtin_printf(" Set pins: CLK_PIN=%d CMD_PIN=%d DAT_PIN=%d POW_PIN=%d LED_PIN=%d\n",
                     PIN_CLK, PIN_CMD, PIN_DAT0, PIN_POW, PIN_LED);
#endif    
    clkdivider = CLK_DIV;
    clkpin = PIN_CLK;
    cmdpin = PIN_CMD;
    dat0pin = PIN_DAT0;
    if( PIN_POW < 0 )    // power switch pin (-1 == not used)
        PIN_POW = PIN_CLK;    // use clock smartpin as dummy out
    if( PIN_LED < 0 )    // activity LED pin (-1 == not used)
        PIN_LED = PIN_CLK;    // use clock smartpin as dummy out
    powpin = PIN_POW;
    ledpin = PIN_LED << 8;    // doubles up as SDSC byte addressing flag in bit0
                              // tripples up as SDHC caching flag in bit1
    return RES_OK;
}


//
// new routine: deinitialize (clean up pins, if necessary)
//
DSTATUS disk_deinitialize( 
    BYTE drv )
{
    if( drv ) {
#ifdef SD_DEBUG
        __builtin_printf( " Deinitialize: bad drv %d\n", drv );
#endif
        return RES_NOTRDY;
    }
#ifdef SD_DEBUG
    __builtin_printf( " Clear pins: %d %d %d %d %d\n", clkpin, cmdpin, dat0pin, powpin, ledpin >> 8 );
#endif
    releasepins();
    _waitms(1);

    return RES_OK;
}


//-----------------------------------------------------------------------
/*
 * Interface to FlexProp VFS
 */
#include <errno.h>
#include <sys/types.h>
#include <sys/vfs.h>
#include <stdbool.h>

#define BLOCK_SIZE  512
#define BLOCK_SHIFT 9
#define BLOCK_MASK  0x1ff

off_t curpos;
uint64_t f_pinmask;

ssize_t v_do_io(vfs_file_t *fil, void *buf_p, size_t count, bool is_write)
{
    unsigned char *buf = (unsigned char *)buf_p;
    unsigned startoff = curpos & BLOCK_MASK;
    unsigned blocks;
    ssize_t  bytes_io = 0;
    unsigned lba = curpos >> BLOCK_SHIFT;
    int res;
    
    if (startoff) {
        /* we have to do the I/O for the first sector */
        /* for now throw up our hands and punt, we don't support arbitrary seeks */
        return -1;
    }
    blocks = count >> BLOCK_SHIFT;
    if (is_write) {
        res = disk_write(0, buf, lba, blocks);
    } else {
        res = disk_read(0, buf, lba, blocks);
    }
    if (res == RES_OK) {
        unsigned update = blocks << BLOCK_SHIFT;
        bytes_io += update;
        count -= update;
        curpos += update;
    }
    if (count) {
        // handle the trailing bytes;
        // not supported for now, I/O must be on block boundaries
    }
    return bytes_io;
}

ssize_t v_read(vfs_file_t *fil, void *buf, size_t count)
{
    return v_do_io(fil, buf, count, false);
}
ssize_t v_write(vfs_file_t *fil, void *buf, size_t count)
{
    return v_do_io(fil, buf, count, true);
}

int v_ioctl(vfs_file_t *fil, int arg, void *buf)
{
    DRESULT res;
    res = disk_ioctl(0, arg, buf);
    if (res)
        return _seterror(EINVAL);
    return 0;
}

off_t v_lseek(vfs_file_t *fil, off_t off, int whence)
{
    if (whence == 0) {
        curpos = off;
    } else if (whence == 1) {
        curpos += off;
    } else {
        curpos = -off;
    }
    return curpos;
}

int v_flush(vfs_file_t *fil)
{
    return 0;
}

int v_close(vfs_file_t *fil)
{
    disk_deinitialize(0);
    _freepins(f_pinmask);
    return 0;
}

int v_putc(int c, vfs_file_t *fil) {
    if (v_write(fil, &c, 1) == 1) return c;
    return -1;
}
int v_getc(vfs_file_t *fil) {
    int c = 0;
    if (v_read(fil, &c, 1) == 1) return c;
    return -1;
}
